@@ -6,8 +6,9 @@ import (
 	"net/http"
 	"net/http/pprof"
 	"runtime"
+	"time"
 
-	"github.com/sirupsen/logrus"
+	"github.com/moby/buildkit/util/bklog"
 	"golang.org/x/net/trace"
 )
 
@@ -24,7 +25,7 @@ func setupDebugHandlers(addr string) error {
 
 	m.Handle("/debug/gc", http.HandlerFunc(func(rw http.ResponseWriter, req *http.Request) {
 		runtime.GC()
-		logrus.Debugf("triggered GC from debug endpoint")
+		bklog.G(req.Context()).Debugf("triggered GC from debug endpoint")
 	}))
 
 	// setting debugaddr is opt-in. permission is defined by listener address
@@ -36,7 +37,16 @@ func setupDebugHandlers(addr string) error {
 	if err != nil {
 		return err
 	}
-	logrus.Debugf("debug handlers listening at %s", addr)
-	go http.Serve(l, m)
+	server := &http.Server{
+		Addr:              l.Addr().String(),
+		Handler:           m,
+		ReadHeaderTimeout: time.Minute,
+	}
+	bklog.L.Debugf("debug handlers listening at %s", addr)
+	go func() {
+		if err := server.Serve(l); err != nil {
+			bklog.L.Errorf("failed to serve debug handlers: %v", err)
+		}
+	}()
 	return nil
 }
